package jodie.core;

import java.io.FileInputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;

import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/*
 * Created on 1-dic-2004
 */

public class Config {

	// isWin is used to autodetect OS 
	// Now I use them just to set the correct params for my windows/linux 
	// installations, but they can be used for higher purposes
	private static final String 	OS_NAME = System.getProperty("os.name").toLowerCase();
	private static final boolean 	isWin = OS_NAME.indexOf("windows")>=0 ? true:false;
	
	public  static final String		cfgXml		= "./config.xml";

	// -------------------------------------------------------------------------
	// BASE PARAMS
	// -------------------------------------------------------------------------
	// DEBUG
	public static boolean		DEBUG			= true;
	// HTTP settings
	public static int		httpPort		= 8080;
	// GUI settings
	public static boolean		hasGUI			= false;
	// Graphviz settings: set to false if the app is not installed
	public static boolean		hasGV			= false;
	

	// -------------------------------------------------------------------------
	// OUTPUT PARAMS
	// -------------------------------------------------------------------------
	// Most of the following params will get default values. These are used 
	// if the cfg file is not found, but can be overriden both by config file
	// params or command line params (at least, this is how it should work)
	public static		String	tableauxXml		= "tableaux.xml";		
	public static		String	dotFileName		= "mala.dot";
	public static		String	dotOutput		= "png";
	public static		String	pngFileName		= "mala.png";
	
	// -------------------------------------------------------------------------
	// PATH PARAMS (these might change, depending on the OS)
	// -------------------------------------------------------------------------
	public static		String	dataPath;
	public static 		String	dotHomeDir;

	// -------------------------------------------------------------------------
	// TEMP PARAMS (these will be deleted, sooner or later)
	// -------------------------------------------------------------------------
	public static		String	tellsXml		= "pizza.xml"; 
	//public static		String	tellsXml		= "parenti.xml"; 

	// -------------------------------------------------------------------------
	// FIXED PARAMS (these won't change and won't be saved in cfg file)
	// -------------------------------------------------------------------------
	public static		String[] dotCommand		= new String[3]; 

	
	// window settings (these vars don't change during runtime)
	public static final int 	windowWidth		= 640;
	public static final int 	windowHeight		= 480;
	public static final String	windowTitle		= "Java Owl-DL Inference Engine";

	// strip away namespace to make names shorter 
	public static final boolean niftyNames		= true;

	// LOG settings
	public static final String	log_main_name	= "Jodie"; 
	public static final String	log_dig_name	= "DIG req"; 
	public static final String	log_res_name	= "DIG res"; 
	public static final String	log_kb_name		= "KB"; 
	public static final String	log_tab_name	= "Tab"; 

	public static		int		log_main_type	= Log.GUI; 
	public static		int		log_dig_type	= Log.GUI;
	public static		int		log_res_type	= Log.GUI;
	public static		int		log_kb_type		= Log.CMD;
	public static		int		log_tab_type	= Log.CMD;

	// Strategies settings: we use a hash to map expressivity 
	// with strategy, setting an "ALL" fallback value if a strategy
	// is not found
	public static HashMap		strategies		= new HashMap(); 

	// Configuration parameters: they are used to load the cfg
	// from a file and then assign the correct values to the
	// public params
	private static HashMap		cfgParams		= new HashMap();
	
	// core
	public static Core			core;
	
	public Config (Core c, String[] args){
		core = c;

		// set strategies 
		strategies.put("ALL","SHf");

		// set up config 
		dataPath = isWin ?
				"D:\\incomin\\poli\\thesis\\data\\": 	// Windoze
				"/home/eynard/data/";					// *NIX
		
		dotHomeDir = isWin ?
				"C:\\Programmi\\ATT\\Graphviz\\bin":	// Windoze
				"/home/eynard/graphviz/bin";			// *NIX

		// read configuration file (overridden by cmdline)
		updateCfg(cfgXml);

		// (TEMPORARY) INPUT FILE(S)
		tellsXml 	= dataPath+tellsXml;

		// OUTPUT FILE(S)
		tableauxXml	= dataPath+tableauxXml;
		dotFileName	= dataPath+dotFileName;
		pngFileName	= dataPath+pngFileName;
		
		if (isWin){
			dotCommand[0] = "cmd";
			dotCommand[1] = "/c";
			dotCommand[2] = "dot -T "+ dotOutput + " " + dotFileName + " -o " + pngFileName;
		}else{
			dotCommand[0] = "/bin/sh";
			dotCommand[1] = "-c";
			dotCommand[2] = "./dot -T "+ dotOutput + " " + dotFileName + " -o " + pngFileName;
		}
		
		// read command line args (they override config file)
		for (int i=0; i<args.length; i++){
			if (args[i].equalsIgnoreCase("nogui")){
				hasGUI = false;
			}
		}
	} 

// - MAIN METHOD ENDS HERE --------------------------------------------------------
	
	public String toString(){
		String s ="";
		s+= "DEBUG: " + DEBUG;
		s+= "httpPort: " + httpPort;
		s+= "hasGUI: " + hasGUI;
		s+= "hasGV: " + hasGV;
		s+= "tableauxXml: " + tableauxXml;		
		s+= "dotFileName: " + dotFileName;
		s+= "dotOutput: " + dotOutput;
		s+= "pngFileName: " + pngFileName;
		s+= "dataPath: " + dataPath;
		s+= "dotHomeDir: " + dotHomeDir;
		s+= "tellsXml: " + tellsXml;
		return s;
	}
	
	private void updateCfg(String filename){
		HashMap hm;
		cfgParams = parseCfgFile (cfgXml);
		
		// update local configuration variables, depending on 
		// cfgParams hash values loaded from the file
		hm = (HashMap)cfgParams.get("BASE");
		DEBUG		= ((String)hm.get("DEBUG")).equalsIgnoreCase("true") ? true : false;
		httpPort	= (Integer.parseInt((String)hm.get("HTTPPORT")));
		hasGUI		= ((String)hm.get("HASGUI")).equalsIgnoreCase("true") ? true : false;
		hasGV		= ((String)hm.get("HASGV")).equalsIgnoreCase("true") ? true : false;

		hm = (HashMap)cfgParams.get("LOG");
		tableauxXml	= (String)hm.get("TABLEAUXXML");		

		log_main_type	= (Integer.parseInt((String)hm.get("MAIN"))); 
		log_dig_type	= (Integer.parseInt((String)hm.get("DIG")));
		log_res_type	= (Integer.parseInt((String)hm.get("RES")));
		log_kb_type		= (Integer.parseInt((String)hm.get("KB")));
		log_tab_type	= (Integer.parseInt((String)hm.get("TAB")));
		
		hm = (HashMap)cfgParams.get("OUTPUT");
		tableauxXml	= (String)hm.get("TABLEAUXXML");		
		dotFileName	= (String)hm.get("DOTFILENAME");
		dotOutput	= (String)hm.get("IMAGEFORMAT");
		pngFileName	= (String)hm.get("IMGFILENAME");

		hm = (HashMap)cfgParams.get("PATH");
		dataPath	= (String)hm.get("DATAPATH");
		dotHomeDir	= (String)hm.get("DOTHOMEDIR");

		hm = (HashMap)cfgParams.get("TEMP");
		tellsXml	= (String)hm.get("TELLSXML");

		hm = (HashMap)cfgParams.get("STRATEGIES");
		Iterator hi = hm.keySet().iterator();
		while (hi.hasNext()){
			String key = (String)hi.next();
			strategies.put(key,(String)hm.get(key));
		}
	}
	
	/* 
	 * loads configuration file, parses it and builds a hashmap
	 * containing configuration params
	 */
	private HashMap parseCfgFile (String filename){
		HashMap config = new HashMap();

		if (DEBUG) System.out.print ("Reading configuration params from " + filename + "... ");
		try {
			DOMParser parser = new DOMParser();
			FileInputStream in = new FileInputStream(filename); 
			parser.parse(new InputSource(in));

			// create the root node
			Node root = parser.getDocument().getChildNodes().item(0);

			// work on children: they should all be named "profile"
			LinkedList children = getChildren (root,"profile");
			for (int i=0; i<children.size(); i++){
				// get section name (common, nix, win)
				Node child = (Node)children.get(i);
				String sectName = child.getAttributes().getNamedItem("name").getNodeValue();
				if (sectName.equalsIgnoreCase("common") ||
					(sectName.equalsIgnoreCase("win") && isWin) ||
					(sectName.equalsIgnoreCase("nix") && !isWin)
						){

					// save common params
					LinkedList nephews = getChildren(child,null);
					for (int j=0;j<nephews.size();j++){
						Node nephew = (Node)nephews.get(j);
						HashMap hm = new HashMap();
						LinkedList params = getChildren(nephew,null);
						for (int k=0;k<params.size();k++){
							Node param = (Node)params.get(k);
							hm.put(param.getLocalName().toUpperCase(),param.getChildNodes().item(0).getNodeValue());
						}
						config.put(nephew.getLocalName().toUpperCase(),hm);
					}
				}
			}
			if (DEBUG) System.out.println ("done.");
		} catch (Exception e){
			System.out.println (e);
			System.exit(0);
		}
		
		return config;
	}

	// this returns only DOM child nodes with a given NodeName
	private LinkedList getChildren (Node n, String name){
		LinkedList nl = new LinkedList();

		NodeList childnodes = n.getChildNodes();
		for (int i=0;i<childnodes.getLength();i++){
			Node child = childnodes.item(i);
			if (child.getNodeType()==1){
				if (name == null || child.getNodeName().equalsIgnoreCase(name))
					nl.add(child);
			}
		}
		return nl;
	} 
}
