/*
 * Created on Feb 24, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jodie.dig;

import jodie.core.Config;
import jodie.core.Log;
import jodie.kb.KnowledgeBase;
import jodie.kb.Term;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 * @author eynard
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class DIG {

	private boolean DEBUG = Config.DEBUG || true;

	// DIG TOKENS DEFINITIONS 
	// request tags (and mode names)
	private static final String IDENT			= "getIdentifier";
	private static final String NEWKB 			= "newkb";
	private static final String RELKB 			= "releaseKB";
	private static final String CLRKB 			= "clearKB";
	private static final String TELLS 			= "tells";
	private static final String ASKS 			= "asks";
	private static final String NONE 			= "";
	
	// concept language
	private static final String TOP 			= "top";
	private static final String BOTTOM			= "bottom";						
	private static final String CATOM			= "catom";

	private static final String AND 			= "and";
	private static final String OR	 			= "or";
	private static final String NOT 			= "not";

	private static final String SOME			= "some";
	private static final String ALL 			= "all";
	private static final String ATM				= "atmost";
	private static final String ATL				= "atleast";
	private static final String ISET			= "iset"; // individual set (enumerazione)

	private static final String DEFINED			= "defined";
	private static final String STRINGMIN		= "stringmin";
	private static final String STRINGMAX		= "stringmax";
	private static final String STRINGEQUALS	= "stringequals";
	private static final String STRINGRANGE		= "stringrange";
	private static final String INTMIN			= "intmin";
	private static final String INTMAX			= "intmax";
	private static final String INTEQUALS		= "intequals";
	private static final String INTRANGE		= "intrange";

	private static final String RATOM			= "ratom";
	private static final String FEATURE			= "feature";
	private static final String INVERSE			= "inverse";
	private static final String ATTRIBUTE		= "attribute";
	private static final String CHAIN			= "chain";

	private static final String INDIVIDUAL		= "individual";

	// tell language
	private static final String DEFCONCEPT		= "defconcept";
	private static final String DEFROLE			= "defrole";
	private static final String DEFFEATURE		= "deffeature";
	private static final String DEFATTRIBUTE	= "defattribute";
	private static final String DEFINDIVIDUAL 	= "defindividual";

	private static final String IMPLIESC		= "impliesc";
	private static final String EQUALC			= "equalc";
	private static final String DISJOINT		= "disjoint";

	private static final String IMPLIESR		= "impliesr";
	private static final String EQUALR			= "equalr";
	private static final String DOMAIN			= "domain";
	private static final String RANGE			= "range";
	private static final String RANGEINT		= "rangeint";
	private static final String RANGESTRING		= "rangestring";
	private static final String TRANSITIVE		= "transitive";
	private static final String FUNCTIONAL		= "functional";

	private static final String INSTANCEOF		= "instanceof";
	private static final String RELATED			= "related";
	private static final String VALUE			= "value";

	// attributes
	private static final String NAME			= "name";
	private static final String NUM				= "num";
	
	private KnowledgeBase kb;
	private Res res;
	
	public DIG() {
		kb = Config.core.getKB();
	}

	public void parse(Document document, Res res)throws Exception{
		this.res = res;

		// get child nodes
		NodeList childnodes = document.getChildNodes();
		
		// check the number of children
		int childnum = childnodes.getLength();
		if (childnum!=1){
			// if document has more than one child, send an error
			res.makeErr(Res.ERR_MALFORMED_REQUEST);
			return;
		}

		Node child = childnodes.item(0);
		// ok, I have one child here... let's match it with its 
		// function, depending on its name
		if (child.getLocalName().equalsIgnoreCase(IDENT)){
			parseIdent(child);
		}else if (child.getLocalName().equalsIgnoreCase(NEWKB)){
			parseNewKB(child);
		}else if (child.getLocalName().equalsIgnoreCase(RELKB)){
			parseRelKB(child);
		}else if (child.getLocalName().equalsIgnoreCase(TELLS)){
			parseTells(child);
		}else if (child.getLocalName().equalsIgnoreCase(ASKS)){
			parseAsks(child);
		}else{
			// create an error response
			res.makeErr(Res.ERR_USUPPORTED_OPERATION);
			return;
		}
	}

	private void parseIdent (Node n){
		res.makeIdent();
	}
	
	private void parseNewKB (Node n){
		res.makeNewKB();
	}
	
	private void parseRelKB (Node n){
		res.makeErr();
	}

	private void parseTells (Node n){
		// check if we just got a clearKB request
		if (n.getChildNodes().item(1).getNodeName().equalsIgnoreCase(CLRKB)){
			res.makeOk();
			return;
		}

		// use an xmlbean to manage Tells content
		try {
			Config.core.timers.startTimer("DIG Parsing");
			
			// used later for concept axioms (uses node parsing instead of xmlbeans)
			NodeList nol, nel; 
			int i, j;
			
			// first, get all the definitions
			// DEFCONCEPT
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(DEFCONCEPT)){
					String cname = child.getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Adding concept " + cname);
					kb.addDefConcept(cname);
				}
			}
			
			// DEFROLE
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(DEFROLE)){
					String rname = child.getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Adding role " + rname);
					kb.addDefRole(rname);
				}
			}

			// DEFFEATURE
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(DEFFEATURE)){
					String fname = child.getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Adding feature " + fname);
					kb.addDefFeature(fname);
				}
			}

			// DEFATTRIBUTE
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(DEFATTRIBUTE)){
					String aname = child.getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Adding attribute " + aname);
					kb.addDefAttribute(aname);
				}
			}

			// DEFINDIVIDUAL
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(DEFINDIVIDUAL)){
					String iname = child.getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Adding individual " + iname);
					kb.addDefIndividual(iname);
				}
			}

			// then, save role axioms

			// IMPLIESR
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(IMPLIESR)){
					// node[1] is an ratom whose "NAME" attribute I'm interested in
					String role1 = child.getChildNodes().item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					
					// node[3] is an <inverse> tag, which has an ratom child 
					//         (item[1]) whose "NAME" attribute I'm interested in
					String role2 = child.getChildNodes().item(3).getAttributes().getNamedItem(NAME).getNodeValue();

					if (DEBUG) Config.core.log.out("Setting Implies for role " + role1 + " (" + role2 + ")"); 
					kb.addImpliesR(role1,role2);
				}
			}

			
			// EQUALR
			// NOTE: equalr is usually used to specify inverse roles (role,inverse(role))
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(EQUALR)){
					// node[1] is an ratom whose "NAME" attribute I'm interested in
					String role1 = child.getChildNodes().item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					
					// node[3] is an <inverse> tag, which has an ratom child 
					//         (item[1]) whose "NAME" attribute I'm interested in
					String role2 = child.getChildNodes().item(3).getChildNodes().item(1).getAttributes().getNamedItem(NAME).getNodeValue();

					if (DEBUG) Config.core.log.out("Setting Inverse for role " + role1 + " (" + role2 + ")"); 
					kb.addInverseR(role1,role2);
				}
			}
			
			// DOMAIN
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(DOMAIN)){
					// node[1] is an ratom whose "NAME" attribute I'm interested in
					String role = child.getChildNodes().item(1).getAttributes().getNamedItem(NAME).getNodeValue();

					String s = "Setting Domain(s) for role " + role + ":"; 

					// node[3] is an <or> tag, which might have various catom children 
					nel = child.getChildNodes().item(3).getChildNodes();
					for (j=0; j<nel.getLength(); j++){
						Node nephew = nel.item(j);
						if (nephew.getNodeName().equalsIgnoreCase(CATOM)){
							String domain = nephew.getAttributes().getNamedItem(NAME).getNodeValue();
							s += " " + domain;
							kb.setRoleDomain(role,domain);
						}
					}
					if (DEBUG) Config.core.log.out(s);
				}
			}

			// RANGE
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(RANGE)){
					// node[1] is an ratom whose "NAME" attribute I'm interested in
					String role = child.getChildNodes().item(1).getAttributes().getNamedItem(NAME).getNodeValue();

					String s = "Setting Range(s) for role " + role + ":"; 

					// node[3] is an <or> tag, which might have various catom children 
					nel = child.getChildNodes().item(3).getChildNodes();
					for (j=0; j<nel.getLength(); j++){
						Node nephew = nel.item(j);
						if (nephew.getNodeName().equalsIgnoreCase(CATOM)){
							String range = nephew.getAttributes().getNamedItem(NAME).getNodeValue();
							s += " " + range;
							kb.setRoleRange(role,range);
						}
					}
					if (DEBUG) Config.core.log.out(s);
				}
			}
			
			// RANGEINT
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(RANGEINT)){
					// node[1] is an attribute whose "NAME" attribute I'm interested in
					String attribute = child.getChildNodes().item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Setting RangeInt for attribute " + attribute);
					kb.setRangeInt(attribute);
				}
			}
	
			// RANGESTRING
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(RANGESTRING)){
					// node[1] is an attribute whose "NAME" attribute I'm interested in
					String attribute = child.getChildNodes().item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Setting RangeString for attribute " + attribute);
					kb.setRangeString(attribute);
				}
			}

			// TRANSITIVE
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(TRANSITIVE)){
					// node[1] is an ratom whose "NAME" attribute I'm interested in
					String role = child.getChildNodes().item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Setting Transitive for role " + role);
					kb.setTransitive(role);
				}
			}

			// FUNCTIONAL
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(FUNCTIONAL)){
					// node[1] is an ratom whose "NAME" attribute I'm interested in
					String role = child.getChildNodes().item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Setting Functional for role " + role);
					kb.setFunctional(role);
				}
			}

			// then, save individual axioms

			/**
			 * 
			 * TODO add individual axioms here!
			 *		related (is added here, needs function body inside KB)
			 *		value
			 *		==> The ones below here are still in BEANS format, convert them  
			 */
			
/*			
			// instanceof axioms are all in the form i,c (i is an instance of c)
			Instanceof[] inst = tells.getInstanceofArray();
			for (i=0; i<inst.length; i++){
				if (DEBUG) Config.core.log.out("Setting InstanceOf for Individual " + inst[i].getIndividual().getName() 
						+ "(" + inst[i].getCatom().getName() + ")");
				kb.addInstanceOf(inst[i].getIndividual().getName(),inst[i].getCatom().getName());
			}
			
			// related axioms are all in the form i,r,i
			Related[] rel = tells.getRelatedArray();
			for (i=0; i<rel.length; i++){
				if (DEBUG) Config.core.log.out("Setting Related for Individual " + rel[i].getIndividualArray(0).getName() 
						+ "(" + rel[i].getRatom().getName() + ", "+ rel[i].getIndividualArray(1).getName() +")");
				kb.addRelated(rel[i].getIndividualArray(0).getName(),rel[i].getRatom().getName(),rel[i].getIndividualArray(1).getName());
			}
*/		
			
			// finally, save concept axioms
			// NOTE: to better manage to-term conversion, concept axioms are not parsed
			//       with beans, but with DOM Nodes: be careful, if somehting changes
			//       you'll probably have to give a look HERE!

			// IMPLIESC
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(IMPLIESC)){
					nel = child.getChildNodes();
					// NOTE: node[1] and node[3] are the nodes I'm interested in
					//       (the other ones are just #text fields)
					
					// the first one is always a catom (right?)
					String cpt1 = nel.item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					//if (DEBUG) Config.core.log.out("Adding Implies for concept " + cpt1);
					
					// the second one is usually a complex concept, we'll have
					// to convert it to a Term before we save it
					Term cpt2 = createTerm(nel.item(3));
					if (DEBUG) Config.core.log.out(cpt1 + " => " + cpt2);
					kb.addImpliesC (cpt1,cpt2);
				}
			}

			// EQUALC
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(EQUALC)){
					nel = child.getChildNodes();
					// NOTE: node[1] and node[3] are the nodes I'm interested in
					//       (the other ones are just #text fields)
					
					// the first one is always a catom (right?)
					String cpt1 = nel.item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Adding Equals for concept " + cpt1);
					
					// the second one is usually a complex concept, we'll have
					// to convert it to a Term before we save it
					Term cpt2 = createTerm(nel.item(3));
					if (DEBUG) Config.core.log.out(cpt1 + " == " + cpt2);
					if (cpt2 != null) kb.addEqualC (cpt1,cpt2);
				}
			}

			// DISJOINT
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(DISJOINT)){
					nel = child.getChildNodes();
					// NOTE: node[1] and node[3] are the nodes I'm interested in
					//       (the other ones are just #text fields)
					
					// the first one is always a catom? NO! IT CAN BE AN ISET!
					if (nel.item(1).getLocalName() != CATOM){
						if (DEBUG) Config.core.log.out("Found an ISET inside a DISJOINT: it hasn't been implemented yet...");
						 break;
					}
					
					// catom code follows (iset has to be implemented yet...)
					String cpt1 = nel.item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					if (DEBUG) Config.core.log.out("Adding Disjoint for concept " + cpt1);
					
					// the second one is usually a complex concept, we'll have
					// to convert it to a Term before we save it
					Term cpt2 = createTerm(nel.item(3));
					if (DEBUG) Config.core.log.out(cpt1 + " != " + cpt2);
					kb.addDisjoint (cpt1,cpt2);
				}
			}
			Config.core.timers.stopTimer("DIG Parsing");

			// SHOW THE WHOLE KB!!!
			Log log = new Log (Config.log_kb_name,Config.log_kb_type);
			log.out(kb.toString());
			log.out("KB has the expressivity of "+ kb.getEx().toString());
			
			// ALCF
			//String 	s = "and(ElenaPizza,some(hasIngredient,some(hasSpiciness,Hot)))"; // VAAAAAAAAAAAAAH! => ALCf!!!
			//String 	s = "and(ElenaPizza,SpicyPizza)"; // VAAAAAAAAAAAAAH! => ALCf!!!

			// -SH---------------------------------------------------------------------
			//String 	s = "and(some(Padre,some(Madre,some(Padre,Persona))),not(some(Antenato,Persona)))"; // VA -> NS
			//String 	s = "and(some(Antenato,Persona),not(some(Antenato,all(Antenato,not(Persona)))))"; // VA -> S
			//String 	s = "and(MargheritaPizza,some(hasTopping,TomatoTopping))"; // VA -> S
			//String 	s = "and(Pizza,not(some(hasBase,PizzaBase)))"; // VA -> CLASH IN UNFOLDING
			//String 	s = "and(ElenaPizza,some(hasTopping,MeatTopping))"; // VA -> NS
			//String 	s = "and(ElenaPizza,VegetarianPizza)"; // VA! -> S 
			//*/String 	s = "and(or(ElenaPizza,MargheritaPizza,SohoPizza),NonVegetarianPizza)"; // VA -> NS
			//String  s = "and(ElenaPizza,or(VegetarianPizza,NonVegetarianPizza))"; *** _DEVE_ ANDARE!!!
			// NOTE: this uses role hierarchies... and it works!
			//       BUT: it doesn't work with standard pizza ontology, because
			//            VegetarianPizza definition is not correct with role hierarchies
			//String 	s = "and(VegetarianPizza,some(hasIngredient,MeatTopping))"; // VA -> S ???

			// -SH(f)------------------------------------------------------------------
			String 	s = "and(MargheritaPizza,some(hasTopping,ZucchineTopping))"; // VAAAAAAAAAAAAAH! => NS!
			
			boolean satis = kb.isSatisfiable(s);
			System.out.println ("Tableaux " + (satis ? "is satisfiable" : "is not satisfiable"));
			
			// return ok 
			res.makeOk();
		} catch (Exception e) {
			e.printStackTrace();
			res.makeErr();
			return;
		}
	}

	private void parseAsks  (Node n){
		res.makeOk();
	}

	/*	
	 * createTerm creates a Term starting from a Node type
	 */	
	public Term createTerm(Node n){ 
		if (n==null) return null;
		
		Term t;
		String nm = n.getNodeName();
		
		// if name is ratom or catom then create a simple term
		if (nm.equalsIgnoreCase(CATOM)){
			t = new Term(Term.TERM,n.getAttributes().getNamedItem(NAME).getNodeValue());
		}
		
		else if (nm.equalsIgnoreCase(RATOM)){
			t = new Term(Term.TERM,Term.ROLE,n.getAttributes().getNamedItem(NAME).getNodeValue());
		}

		// TOP
		else if (nm.equalsIgnoreCase(TOP)){
			t = Term.TOP;
		}

		// else switch
		else if (nm.equalsIgnoreCase(AND) ||
				 nm.equalsIgnoreCase(OR)  ||
				 nm.equalsIgnoreCase(NOT) ||
				 nm.equalsIgnoreCase(SOME)||
				 nm.equalsIgnoreCase(ALL)){
			// create a complex term
			t = new Term (Term.OPER,nm);

			// foreach child
			NodeList nl = n.getChildNodes();
			for (int i=0; i<nl.getLength(); i++){
				if (nl.item(i).getNodeType()==Node.ELEMENT_NODE){
					t.addChild(createTerm(nl.item(i)));
				}
			}
		}

		// else switch (extract num from restrictions)
		else if (nm.equalsIgnoreCase(ATL)||
				 nm.equalsIgnoreCase(ATM)){

			// create a complex term
			t = new Term (Term.OPER,nm);
			
			// add a restriction value child
			t.setVal(Integer.parseInt(n.getAttributes().getNamedItem(NUM).getNodeValue()));

			// foreach child
			NodeList nl = n.getChildNodes();
			for (int i=0; i<nl.getLength(); i++){
				if (nl.item(i).getNodeType()==Node.ELEMENT_NODE){
					t.addChild(createTerm(nl.item(i)));
				}
			}
			// set "N" expressivity flag (non-qualified restrictions) 
			kb.getEx().isN(true);
		}

		else {
			System.out.println ("***ERROR IN TERM CREATION: unknown term " + nm);
			t = new Term(Term.TERM,"null");
		}
		
		return t;
	}
}
