/*
 * Created on Feb 24, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jodie.dig;

import java.io.FileWriter;
import java.io.IOException;

import jodie.core.Config;
import jodie.core.Log;
import jodie.core.Utils;
import jodie.kb.KnowledgeBase;
import jodie.kb.Term;
import jodie.tableaux.Assertion;
import jodie.tableaux.Tableaux;

import org.apache.xmlbeans.XmlException;
import org.kr.dl.dig.v1_1.AttributeType;
import org.kr.dl.dig.v1_1.Named;
import org.kr.dl.dig.v1_1.Role;
import org.kr.dl.dig.v1_1.RoleConceptPair;
import org.kr.dl.dig.v1_1.RolePair;
import org.kr.dl.dig.v1_1.TellsDocument;
import org.kr.dl.dig.v1_1.InstanceofDocument.Instanceof;
import org.kr.dl.dig.v1_1.RelatedDocument.Related;
import org.kr.dl.dig.v1_1.TellsDocument.Tells;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 * @author eynard
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class DIG_beans {

	private boolean DEBUG = Config.DEBUG || true;

	// DIG TOKENS DEFINITIONS 
	// request tags (and mode names)
	private static final String IDENT			= "getIdentifier";
	private static final String NEWKB 			= "newkb";
	private static final String RELKB 			= "releaseKB";
	private static final String CLRKB 			= "clearKB";
	private static final String TELLS 			= "tells";
	private static final String ASKS 			= "asks";
	private static final String NONE 			= "";
	
	// concept language
	private static final String TOP 			= "top";
	private static final String BOTTOM			= "bottom";						
	private static final String CATOM			= "catom";
	private static final String RATOM			= "ratom";

	private static final String AND 			= "and";
	private static final String OR	 			= "or";
	private static final String NOT 			= "not";

	private static final String SOME			= "some";
	private static final String ALL 			= "all";
	private static final String ATM				= "atmost";
	private static final String ATL				= "atleast";
	private static final String ISET			= "iset"; // individual set (enumerazione)

	// tell language
	private static final String IMPLIESC		= "impliesc";
	private static final String EQUALC			= "equalc";
	private static final String DISJOINT		= "disjoint";

	// attributes
	private static final String NAME			= "name";
	private static final String NUM				= "num";
	
	private KnowledgeBase kb;
	private Res res;
	
	public DIG_beans() {
		kb = Config.core.getKB();
	}

	public void parse(Document document, Res res)throws Exception{
		this.res = res;

		// get child nodes
		NodeList childnodes = document.getChildNodes();
		
		// check the number of children
		int childnum = childnodes.getLength();
		if (childnum!=1){
			// if document has more than one child, send an error
			res.makeErr(Res.ERR_MALFORMED_REQUEST);
			return;
		}

		Node child = childnodes.item(0);
		// ok, I have one child here... let's match it with its bean,
		// depending on its name
		if (child.getLocalName().equalsIgnoreCase(IDENT)){
			parseIdent(child);
		}else if (child.getLocalName().equalsIgnoreCase(NEWKB)){
			parseNewKB(child);
		}else if (child.getLocalName().equalsIgnoreCase(RELKB)){
			parseRelKB(child);
		}else if (child.getLocalName().equalsIgnoreCase(TELLS)){
			parseTells(child);
		}else if (child.getLocalName().equalsIgnoreCase(ASKS)){
			parseAsks(child);
		}else{
			// create an error response
			res.makeErr(Res.ERR_USUPPORTED_OPERATION);
			return;
		}
	}

	private void parseIdent (Node n){
		res.makeIdent();
	}
	
	private void parseNewKB (Node n){
		res.makeNewKB();
	}
	
	private void parseRelKB (Node n){
		res.makeErr();
	}

	private void parseTells (Node n){
		// check if we just got a clearKB request
		if (n.getChildNodes().item(1).getNodeName().equalsIgnoreCase(CLRKB)){
			res.makeOk();
			return;
		}

		// use an xmlbean to manage Tells content
		try {
			Config.core.timers.startTimer("DIG Parsing");
			
			//Config.core.timers.startTimer("DIG Parsing_beans");
			TellsDocument doc = TellsDocument.Factory.parse(n);
			//Config.core.timers.stopTimer("DIG Parsing_beans");
			
			Tells tells = doc.getTells();

			// used later for concept axioms (uses node parsing instead of xmlbeans)
			NodeList nol, nel; 
			int i, j;
			
			// first, get all the definitions
			// DEFCONCEPT
			Named[] concepts = tells.getDefconceptArray();
			for (i=0; i<concepts.length; i++){
				if (DEBUG) Config.core.log.out("Adding concept " + concepts[i].getName());
				kb.addDefConcept(concepts[i].getName());
			}

			// DEFROLE
			Named[] roles = tells.getDefroleArray();
			for (i=0; i<roles.length; i++){
				if (DEBUG) Config.core.log.out("Adding role " + roles[i].getName());
				kb.addDefRole(roles[i].getName());
			}

			// DEFFEATURE
			Named[] features = tells.getDeffeatureArray();
			for (i=0; i<features.length; i++){
				if (DEBUG) Config.core.log.out("Adding feature " + features[i].getName());
				kb.addDefFeature(features[i].getName());
			}

			// DEFATTRIBUTE
			Named[] attributes = tells.getDefattributeArray();
			for (i=0; i<attributes.length; i++){
				if (DEBUG) Config.core.log.out("Adding attribute " + attributes[i].getName());
				kb.addDefAttribute(attributes[i].getName());
			}

			// DEFINDIVIDUAL
			Named[] individuals = tells.getDefindividualArray();
			for (i=0; i<individuals.length; i++){
				if (DEBUG) Config.core.log.out("Adding individual " + individuals[i].getName());
				kb.addDefIndividual(individuals[i].getName());
			}

			// then, save role axioms

			// IMPLIESR
			RolePair[] implr = tells.getImpliesrArray();
			for (i=0; i<implr.length; i++){
				if (DEBUG) Config.core.log.out("Setting Implies for role " + implr[i].getRatomArray(0).getName()
						+ "(" + implr[i].getRatomArray(1).getName() + ")"); 
				kb.addImpliesR(implr[i].getRatomArray(0).getName(),implr[i].getRatomArray(1).getName());
			}
			
			// EQUALR
			// NOTE: equalr is usually used to specify inverse roles (role,inverse(role))
			RolePair[] equalr = tells.getEqualrArray();
			for (i=0; i<equalr.length; i++){
				if (DEBUG) Config.core.log.out("Setting Inverse for role " + equalr[i].getRatomArray(0).getName()
						+ " (" + equalr[i].getInverseArray(0).getRatom().getName() + ")"); 
				kb.addInverseR(equalr[i].getRatomArray(0).getName(),equalr[i].getInverseArray(0).getRatom().getName());
			}
			
			// DOMAIN
			RoleConceptPair[] domain = tells.getDomainArray();
			for (i=0; i<domain.length; i++){
				String s = "Setting Domain(s) for role " + domain[i].getRatom().getName() + ":"; 
				Named[] domconc = domain[i].getOr().getCatomArray();
				for (j=0; j<domconc.length; j++){
					s += " " + domconc[j].getName();
					kb.setRoleDomain(domain[i].getRatom().getName(),domconc[j].getName());
				}
				if (DEBUG) Config.core.log.out(s); 
			}

			// RANGE
			RoleConceptPair[] range = tells.getRangeArray();
			for (i=0; i<range.length; i++){
				String s = "Setting Range(s) for role " + range[i].getRatom().getName() + ":"; 
				Named[] rngconc = range[i].getOr().getCatomArray();
				for (j=0; j<rngconc.length; j++){
					s += " " + rngconc[j].getName();
					kb.setRoleRange(range[i].getRatom().getName(),rngconc[j].getName());
				}
				if (DEBUG) Config.core.log.out(s); 
			}
					
			// RANGEINT
			AttributeType[] rangeint = tells.getRangeintArray();
			for (i=0; i<rangeint.length; i++){
				if (DEBUG) Config.core.log.out("Setting RangeInt for attribute " + rangeint[i].getAttribute().getName());
				kb.setRangeInt(rangeint[i].getAttribute().getName());
			}

			// RANGESTRING
			AttributeType[] rangestr = tells.getRangestringArray();
			for (i=0; i<rangestr.length; i++){
				if (DEBUG) Config.core.log.out("Setting RangeString for attribute " + rangestr[i].getAttribute().getName());
				kb.setRangeString(rangestr[i].getAttribute().getName());
			}

			// TRANSITIVE
			Role[] trans = tells.getTransitiveArray();
			for (i=0; i<trans.length; i++){
				if (DEBUG) Config.core.log.out("Setting Transitive for role " + trans[i].getRatom().getName());
				kb.setTransitive(trans[i].getRatom().getName());
			}

			// FUNCTIONAL
			Role[] func = tells.getFunctionalArray();
			for (i=0; i<func.length; i++){
				if (DEBUG) Config.core.log.out("Setting Functional for role " + func[i].getRatom().getName());
				kb.setFunctional(func[i].getRatom().getName());
			}

			
			// then, save individual axioms

			/**
			 * 
			 * TODO add individual axioms here!
			 *		related (is added here, needs function body inside KB)
			 *		value  
			 */
			
			
			// instanceof axioms are all in the form i,c (i is an instance of c)
			Instanceof[] inst = tells.getInstanceofArray();
			for (i=0; i<inst.length; i++){
				if (DEBUG) Config.core.log.out("Setting InstanceOf for Individual " + inst[i].getIndividual().getName() 
						+ "(" + inst[i].getCatom().getName() + ")");
				kb.addInstanceOf(inst[i].getIndividual().getName(),inst[i].getCatom().getName());
			}
			
			// related axioms are all in the form i,r,i
			Related[] rel = tells.getRelatedArray();
			for (i=0; i<rel.length; i++){
				if (DEBUG) Config.core.log.out("Setting Related for Individual " + rel[i].getIndividualArray(0).getName() 
						+ "(" + rel[i].getRatom().getName() + ", "+ rel[i].getIndividualArray(1).getName() +")");
				kb.addRelated(rel[i].getIndividualArray(0).getName(),rel[i].getRatom().getName(),rel[i].getIndividualArray(1).getName());
			}
			
			
			// finally, save concept axioms
			// NOTE: to better manage to-term conversion, concept axioms are not parsed
			//       with beans, but with DOM Nodes: be careful, if somehting changes
			//       you'll probably have to give a look HERE!

			// IMPLIESC
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(IMPLIESC)){
					nel = child.getChildNodes();
					// NOTE: node[1] and node[3] are the nodes I'm interested in
					//       (the other ones are just #text fields)
					
					// the first one is always a catom (right?)
					String cpt1 = nel.item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					//if (DEBUG) Config.core.log.out("Adding Implies for concept " + cpt1);
					
					// the second one is usually a complex concept, we'll have
					// to convert it to a Term before we save it
					Term cpt2 = createTerm(nel.item(3));
					if (DEBUG) Config.core.log.out(cpt1 + " => " + cpt2);
					kb.addImpliesC (cpt1,cpt2);
				}
			}

			// EQUALC
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(EQUALC)){
					nel = child.getChildNodes();
					// NOTE: node[1] and node[3] are the nodes I'm interested in
					//       (the other ones are just #text fields)
					
					// the first one is always a catom (right?)
					String cpt1 = nel.item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					//if (DEBUG) Config.core.log.out("Adding Equals for concept " + cpt1);
					
					// the second one is usually a complex concept, we'll have
					// to convert it to a Term before we save it
					Term cpt2 = createTerm(nel.item(3));
					if (DEBUG) Config.core.log.out(cpt1 + " == " + cpt2);
					kb.addEqualC (cpt1,cpt2);
				}
			}

			// DISJOINT
			nol = n.getChildNodes();
			for (i=0; i<nol.getLength(); i++){
				Node child = nol.item(i);
				if (child.getNodeName().equalsIgnoreCase(DISJOINT)){
					nel = child.getChildNodes();
					// NOTE: node[1] and node[3] are the nodes I'm interested in
					//       (the other ones are just #text fields)
					
					// the first one is always a catom (right?)
					String cpt1 = nel.item(1).getAttributes().getNamedItem(NAME).getNodeValue();
					//if (DEBUG) Config.core.log.out("Adding Equals for concept " + cpt1);
					
					// the second one is usually a complex concept, we'll have
					// to convert it to a Term before we save it
					Term cpt2 = createTerm(nel.item(3));
					if (DEBUG) Config.core.log.out(cpt1 + " != " + cpt2);
					kb.addDisjoint (cpt1,cpt2);
				}
			}
			Config.core.timers.stopTimer("DIG Parsing");

			// SHOW THE WHOLE KB!!!
			Log log = new Log (Config.log_kb_name,Config.log_kb_type);
			log.out(kb.toString());
			log.out("KB has the expressivity of "+ kb.getEx().toString());
			
			String 	s = "and(MargheritaPizza,not(some(hasTopping,TomatoTopping)))";
			Term	t = Utils.stringToTerm(s).getTerm(0);
			
			Config.core.timers.startTimer("Unfolding");
			Term z = t.normalize().simplify();//.unfold(kb.getTBox()).nnf().simplify();
			Config.core.timers.stopTimer("Unfolding");
			
			Config.core.timers.startTimer("Rulez");
			Tableaux tab = new Tableaux(kb);
			tab.getRootNode().addAssertion(new Assertion(z));
			System.out.println (tab);
			tab.applyRules();
			System.out.println (tab);
			Config.core.timers.stopTimer("Rulez");
			
			try {
				FileWriter fw = new FileWriter(Config.tableauxXml);
				fw.write(tab.toXML());
				fw.close();
			} catch (IOException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			
			// return ok 
			res.makeErr();

		} catch (XmlException e) {
			e.printStackTrace();
			res.makeErr();
			return;
		}
	}

	private void parseAsks  (Node n){
		res.makeOk();
	}

	/*	
	 * createTerm creates a Term starting from a Node type
	 */	
	public Term createTerm(Node n){ 

		Term t;
		String nm = n.getNodeName();
		
		// if name is ratom or catom then create a simple term
		if (nm.equalsIgnoreCase(CATOM)){
			t = new Term(Term.TERM,n.getAttributes().getNamedItem(NAME).getNodeValue());
		}
		
		else if (nm.equalsIgnoreCase(RATOM)){
			t = new Term(Term.TERM,Term.ROLE,n.getAttributes().getNamedItem(NAME).getNodeValue());
		}

		// TOP
		else if (nm.equalsIgnoreCase(TOP)){
			t = Term.TOP;
		}

		// else switch
		else if (nm.equalsIgnoreCase(AND) ||
				 nm.equalsIgnoreCase(OR)  ||
				 nm.equalsIgnoreCase(NOT) ||
				 nm.equalsIgnoreCase(SOME)||
				 nm.equalsIgnoreCase(ALL)){
			// create a complex term
			t = new Term (Term.OPER,nm);

			// foreach child
			NodeList nl = n.getChildNodes();
			for (int i=0; i<nl.getLength(); i++){
				if (nl.item(i).getNodeType()==Node.ELEMENT_NODE){
					t.addChild(createTerm(nl.item(i)));
				}
			}
		}

		// else switch (extract num from restrictions)
		else if (nm.equalsIgnoreCase(ATL)||
				 nm.equalsIgnoreCase(ATM)){

			// create a complex term
			t = new Term (Term.OPER,nm);
			
			// add a restriction value child
			t.setVal(Integer.parseInt(n.getAttributes().getNamedItem(NUM).getNodeValue()));

			// foreach child
			NodeList nl = n.getChildNodes();
			for (int i=0; i<nl.getLength(); i++){
				if (nl.item(i).getNodeType()==Node.ELEMENT_NODE){
					t.addChild(createTerm(nl.item(i)));
				}
			}
			// set "N" expressivity flag (non-qualified restrictions) 
			kb.getEx().isN(true);
		}

		else {
			System.out.println ("***ERROR IN TERM CREATION: unknown term " + nm);
			t = new Term(Term.TERM,"null");
		}
		
		return t;
	}
}
