/*
 * Created on 2-feb-2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jodie.kb;

import java.util.HashMap;
import java.util.LinkedList;

import jodie.core.Config;
import jodie.core.Utils;

public class Term implements Cloneable{

	protected boolean DEBUG = Config.DEBUG || true;

	public static final int TERM = 0;
	public static final int OPER = 1;

	public static final int CONCEPT	= 0;
	public static final int ROLE	= 1;

	public static final String AND	= "and";
	public static final String OR	= "or";
	public static final String NOT	= "not";
	public static final String SOME	= "some";
	public static final String ALL	= "all";
	public static final String ATM	= "atmost";
	public static final String ATL	= "atleast";

	// top and bottom
	public static final Term TOP	= new Term(TERM,"TOP");
	public static final Term BOTTOM	= TOP.negate();
	
	// visualization params
	private static final String sep   = ", ";
	private static final String open  = "("; 
	private static final String close = ")"; 
	
	// attributes
	private int type;
	private String name = "";
	private int val = 0; // used for atmost/atleast restrictions
	private int isa = 0; // can be CONCEPT or ROLE
	private boolean unfolded = false;
	
	// this could be instantiated only when I have an OP type...
	private TermList children = new TermList(); 
	
	public Term(int type, String name){
		this (type,CONCEPT,name);
	}

	public Term(int type, int isa, String name){
		this.type = type;
		this.isa  = isa;
		this.name = name;
	}

	public Object clone(){
		Term t = null;
		try {
			t = (Term)super.clone();
			t.type	= type;
			t.name	= name;
			t.val	= val;
			t.isa	= isa;
			t.unfolded = unfolded;
			t.children = (TermList)children.clone();
		} catch (CloneNotSupportedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return t;
	}

	public Term cloneRoot(){
		Term t = new Term (this.type,this.name);
		t.val = this.val;
		t.isa	= isa;
		t.unfolded = unfolded;
		return t;
	}
	
	public boolean isSimple(){
		return type==TERM;
	}

	public boolean isComplex(){
		return type==OPER;
	}

	public boolean isConcept(){
		return isa==CONCEPT;
	}
	
	public boolean isRole(){
		return isa==ROLE;
	}

	public boolean isUnfolded(){
		return this.unfolded;
	}
	
	public void setUnfolded(){
		this.unfolded = true;
	}
	
	public void addChild (Term child){
		children.addLast(child);
	}
	
	public void addChildren (TermList list){
		for (int i=0;i<list.size();i++){
			this.addChild(list.getTerm(i));
		}
	}

	public void setName(String name){
		this.name = name;
	}
	
	public void setVal(int val){
		this.val = val;
	}
	
	public String getName(){
		return this.name;
	} 

	public int getVal(){
		return this.val;
	}

	public String toString(){
		String s;
		
		String name = Utils.niftyName(this.name);
		if (isSimple()){
			s = name;
		}else{
			s = name+open;
			int j = children.size();
			for (int i=0;i<j;i++){
				s += children.get(i).toString();
				if (i<j-1)	s += sep;
			}
			if (name.equalsIgnoreCase(ATL)||name.equalsIgnoreCase(ATM))
				s += ", " + this.val;
			s += close;
		}

		// s += "(isUnfolded="+this.isUnfolded()+")";
		return s;
	}
	
	public Term negate(){
		Term t = new Term(OPER,NOT);
		t.addChild(this);
		return t;
	}
	
	public Term normalize(){
		
		if (this.isSimple()) return this;
		
		if (this.name.equalsIgnoreCase(OR)){
			//System.out.print (this.toString());
			Term and = new Term(OPER,AND);
			and.addChildren(children.negate());
			return and.negate().normalize();
		}
		
		if (this.name.equalsIgnoreCase(SOME)){
			//System.out.print (this.toString());
			Term all = new Term(OPER,ALL);
			// note: for SOME and ALL children are POSITIONAL
			// (the first one is the role and the second is the concept) 
			all.addChild(this.children.getTerm(0));
			all.addChild(this.children.getTerm(1).negate());
			return all.negate().normalize();
		}

		if (this.name.equalsIgnoreCase(ATM)){
			//System.out.print (this.toString());
			Term atl = new Term(OPER,ATL);
			atl.addChild(this.children.getTerm(0));
			atl.addChild(this.children.getTerm(1));
			atl.val = this.val+1; 
			//System.out.println (" => " + not.toString());
			return atl.negate().normalize();
		}
		
		// create a clone of this root
		Term t = this.cloneRoot();
		
		// return a term with every child normalized
		t.addChildren(children.normalize());
		return t;
	}

	public Term lazyUnfold(TBox tbox){
		return lazyUnfold(tbox,new LinkedList());
	}
	
	public Term lazyUnfold(TBox tbox, LinkedList seen){

		Term t1;
		String name = this.name;
		String toString = this.toString();
		Concept c;
		HashMap exp = tbox.getExpansions();

		if (DEBUG) Config.core.log.out ("Lazily unfolding " + this);
		
		// do not unfold roles 
		if (this.isRole()) return this;
		
		// do not unfold concepts which have been unfolded yet
		if (this.isUnfolded()) {
			if (DEBUG) Config.core.log.out ("I won't Lunfold term "+this+", this has been Lunfolded yet");
			return this;
		}

		// if the term has been seen yet do not unfold it
		if (seen.contains(toString)){
			if (DEBUG) Config.core.log.out ("I won't Lunfold term "+this+", this has been seen yet");
			return this;
		} 

		// if the term has been cached yet return its expansion
		if (exp.containsKey(toString)){
			if (DEBUG) Config.core.log.out ("Returning cached value for " + this +": "+ (Term)exp.get(toString));
			return (Term)exp.get(toString);
		}
		
		if (this.isSimple()) {
			if (DEBUG) Config.core.log.out ("Unfolding " + this);
			// we should find the concept in the Tbox
			if ((c=tbox.getConcept(name))!=null){
				// expand the concept 
				seen.addLast(this.toString());
				
				// if (DEBUG) Config.core.log.out ("Seen queue: " + seen);
				
				t1 = c.lazyExpand(tbox,seen);
				// if t1 is null keep the old term
				if (t1==null) t1 = this;

				exp.put(toString,t1);
				//if (DEBUG) Config.core.log.out("t1 has become " + t1);
				seen.removeLast();
				//if (DEBUG) Config.core.log.out("Seen queue has become: " + seen);
			}else{
				if (DEBUG) Config.core.log.out("Could not find concept " + name + " in the TBox");
				t1 = this;
			}
		}else if (this.isComplex()){
			// remember that the function works well on a normalized term
			// so, the only complex terms should be
			// NOT
			// ALL (SOME diventa NOT(ALL(NOT(
			// AND (OR diventa NOT(AND(NOT
			// ... BUT... 
			// sometimes concept expansions return non normalized concepts: CHECK THIS!

			if (	this.name.equalsIgnoreCase(ALL) ||
					this.name.equalsIgnoreCase(SOME)){
//					this.name.equalsIgnoreCase(SOME)||
//					this.name.equalsIgnoreCase(OR)){
				return this;
			}else if ( this.name.equalsIgnoreCase(NOT)){
				// get child
				Term Child = this.getChild(0);
				//if (DEBUG) Config.core.log.out("Unfolding " + this +": child name is " + Child.name);
				if (	Child.name.equalsIgnoreCase(ALL))
//						Child.name.equalsIgnoreCase(AND))
					return this;
			}
			
			// create a clone of the root
			t1 = this.cloneRoot();
			// expand every child
			for (int i=0;i<children.size();i++){
				Term child = children.getTerm(i);
				t1.addChild(child.normalize().lazyUnfold(tbox,seen));
			}
		}else{
			// this should never happen, that is either a term 
			// is simple or it is complex
			Config.core.log.out("Type error for term " + name);
			t1 = null;
		}
		t1.setUnfolded();
		if (DEBUG) Config.core.log.out (this +" became "+ t1);
		return t1;
	} 
	
	public Term nnf(){
		// if you have a simple term just return it 
		if (this.isSimple()) return this;
		
		if (this.name.equalsIgnoreCase(NOT)){
			// get the first (and unique) child
			Term child = this.getChild(0);

			// should not happen, but who knows...
			if (child == null) {
				System.out.println ("What? A NOT without a child?");
				return this;
			}
			
			// if the negation is applied to a term it's OK
			if (child.isSimple()) return this;
			
			// from now on the child is an OPER, let's see its name 
			if (child.name.equalsIgnoreCase(NOT)){
				// double negation, return the first child of "child" in nnf
				return child.getChild(0).nnf();
			}else if (child.name.equalsIgnoreCase(AND)){
				Term t = new Term(OPER,OR);
				t.unfolded = unfolded;
				// nnf and negate every child
				t.addChildren(child.children.negate().nnf());
				return t;
			}else if (child.name.equalsIgnoreCase(OR)){
				Term t = new Term(OPER,AND);
				t.unfolded = unfolded;
				// nnf and negate every child
				t.addChildren(child.children.negate().nnf());
				return t;
			}else if (child.name.equalsIgnoreCase(SOME)){
				Term t = new Term(OPER,ALL);
				t.unfolded = unfolded;
				t.addChild(child.getChild(0));
				t.addChild(child.getChild(1).negate().nnf());
				return t;
			}else if (child.name.equalsIgnoreCase(ALL)){
				Term t = new Term(OPER,SOME);
				t.unfolded = unfolded;
				t.addChild(child.getChild(0));
				t.addChild(child.getChild(1).negate().nnf());
				return t;
			}else if (child.name.equalsIgnoreCase(ATL)){
				Term t = new Term(OPER,ATM);
				t.unfolded = unfolded;
				// the first child is a role
				// the second child is a concept (TOP for OWL)
				t.addChildren(child.getChildren());
				t.val = child.val-1;
				return t;
			}else if (child.name.equalsIgnoreCase(ATM)){
				Term t = new Term(OPER,ATM);
				t.unfolded = unfolded;
				// the first child is a role
				// the second child is a concept (TOP for OWL)
				t.addChildren(child.getChildren());
				t.val = child.val+1;
				return t;
			} else{
				System.out.println ("Term-nnf: Forgot one operator? "+child.name);
				return this;
			}
		}else if ((this.name.equalsIgnoreCase(ATL))||(this.name.equalsIgnoreCase(ATM))){
			return this;
		}else{
			// create a clone of the root
			Term t = this.cloneRoot();
			// nnf every child
			t.addChildren(children.nnf());
			return t;
		}
	}

	public Term simplify(){
		// NOTE: this method works fine on a previously NORMALIZED term
		// first normalize the term
		Term term = this.normalize();
		// Term term = this;
		
		// if you have a simple term just return it 
		if (term.isSimple()) return term;

		// else simplify children
		Term t = term.cloneRoot();
		t.addChildren(term.getChildren().simplify());
		
		if (t.name.equalsIgnoreCase(NOT)){
			Term child = t.getChild(0);
			if(child.isComplex() && child.name.equalsIgnoreCase(NOT)){
				return child.getChild(0);
			}
		}

		if (t.name.equalsIgnoreCase(AND)){
			TermList newChildren = new TermList();

			// if you find a nested AND, push its children up one level
			for (int i=0;i<t.children.size();i++){
				Term child = t.getChild(i);
				if (child.isComplex() && child.name.equalsIgnoreCase(AND)){
					// get all the nephews and add them to tmpCh
					newChildren.add(child.getChildren());
				}else{
					// add the child to tmpCh
					newChildren.add(child);
				}
			}
			
			// at this point, newChildren has to be used instead of children
			// NOTE that all the children should be simplified now

			// strip away dupes, or TOPs
			for (int i=0;i<newChildren.size();i++){
				// get a child
				Term child = newChildren.getTerm(i);
				//System.out.println("Checking if a dupe exists for term " + child);
				if (child.equals(TOP) || newChildren.contains(child,i+1) ){
					newChildren.remove(i);
					i--;
				}
			}
		
			// if there's a clash between children return BOTTOM
			boolean clash = false; 
			for (int i=0;i<newChildren.size()-1;i++){
				// get a child
				Term child = newChildren.getTerm(i);
				// check if there's a clash between children
				if (newChildren.contains(child.negate().simplify(),i+1)){
					clash = true;
					break;
				}
			}
			if (clash){
				// return a simple BOTTOM term
				return BOTTOM;
			}
			
			// if the AND contains a BOTTOM just return BOTTOM
			// (should never happen, but who knows...) 
			if (newChildren.contains(BOTTOM)) return BOTTOM;

			// if AND(ALL(r,c),...,ALL(r,d)) then AND(ALL(r,AND(c,d)),...)
			// scan children to find an ALL
			for (int i=0;i<newChildren.size();i++){
				Term child = newChildren.getTerm(i);
				// if it's an ALL...
				if (child.isComplex()&&child.name.equalsIgnoreCase(ALL)){
					// search another ALL with the same role
					for (int j=i+1;j<newChildren.size();j++){
						Term child2 = newChildren.getTerm(j);
						if (child2.isComplex()&&child2.name.equalsIgnoreCase(ALL)&&
							child2.getChild(0).equals(child.getChild(0))){
							// remove the two ALL 
							newChildren.remove(i);
							newChildren.remove(j-1);
							Term all = new Term(OPER,ALL);
							Term and = new Term(OPER,AND);
							and.addChild(child.getChild(1));
							and.addChild(child2.getChild(1));
							all.addChild(child.getChild(0));
							all.addChild(and.simplify());
							newChildren.add(all);
							i--;
							break;
						}
					}
				}
			}
			
			// do whatever else you want with the term...

			// if, after simplification, AND has only one child 
			// just return it (might happen after ALL collapsing)
			if (newChildren.size()==1) return newChildren.getTerm(0); 
						
			Term t2 = t.cloneRoot();
			t2.addChildren(newChildren);
			return t2;
		}
		
		if (t.name.equalsIgnoreCase(ALL)){
			// check if ALL(role,concept) is ALL (role,TOP)
			Term child = t.getChild(1);
			if (child.equals(TOP)){
				return TOP;
			}
		}
		
		if (t.name.equalsIgnoreCase(ATL)){
			// check if value is 0: ATL(role,0) = TOP
			if (t.val == 0){
				return TOP;
			}
		}
		
		// return the term with all its children simplified 
		return t;
	}

	public boolean equals(Term t){
		if (this.isSimple()) return this.name.equalsIgnoreCase(t.name);
		if (this.val!=t.val) return false;
		return this.name.equalsIgnoreCase(t.name) && this.children.equals(t.children);
	}
	
	public TermList getChildren(){
		return children;
	}
	
	public Term getChild(int i){
		return children.getTerm(i);
	}
}
