/*
 * Created on Mar 22, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jodie.strategies;

import java.util.ListIterator;

import jodie.core.Config;
import jodie.kb.Term;
import jodie.kb.TermList;
import jodie.tableaux.AssertList;
import jodie.tableaux.Assertion;
import jodie.tableaux.Node;
import jodie.tableaux.Tableaux;

/**
 * @author eynard
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public abstract class Strategy {

	protected Tableaux tab	= null;
	protected boolean DEBUG = Config.DEBUG && true;

	public Strategy() {
	}

	public void setTableaux(Tableaux tab){
		this.tab = tab;
	}
	
	public void applyRules(){}
	
	protected void applyUnfoldRule(Node n){
		Config.core.timers.startTimer("[TAB][UNF]");
		// foreach assertion inside the node...
		AssertList assertions = n.getAsserts();
		
		// ... do lazy unfolding
		if (DEBUG) tab.log.out ("[UNF] Executing unfolding on node " + n.getName());
		n.getAsserts().simplify().lazyUnfold(n.getTableaux().getKB().getTBox()).simplify().nnf();
		//n.getAsserts().lazyUnfold(n.getTableaux().getKB().getTBox()).nnf();

		Config.core.timers.stopTimer("[TAB][UNF]");
	}

	protected void applyBCP(Node n){
		Config.core.timers.startTimer("[TAB][BCP]");

		if (DEBUG) tab.log.out("[BCP] Executing applyBCP");
		
		AssertList ass = n.getAsserts();

		ListIterator lj = ass.listIterator();
		// foreach assert...
		while (lj.hasNext()){

			// Get assertion
			Assertion a = (Assertion)lj.next();
			Term t = a.getTerm();
			
			if (t.isComplex() && t.getName().equalsIgnoreCase(Term.OR)){

				//tab.log.out("[BCP] Found an OR in node "+n.getName());
				//tab.log.out("[BCP] Full form is: " + t);

				ListIterator li = t.getChildren().listIterator();
				// foreach concept...
				while (li.hasNext()){
					Term child = (Term)li.next();

					//tab.log.out("[BCP] Checking if child "+child+" clashes");

					// the cases are:
					// simple term or NOT(term): just check for its negation
					if (child.isSimple() || child.getName().equalsIgnoreCase(Term.NOT)){
						//tab.log.out ("[BCP] Child " + child + " is simple");
						if (ass.contains(child.negate().simplify())){
							//tab.log.out("[BCP] Child " + child + " clashes: removing it");
							//tab.log.out("[BCP] => assertions are " + ass);
							//tab.log.out("[BCP] => they contain " + child.negate().simplify());
							li.remove();
						}
					}

					// AND(a,b,c): just check if one of them clashes
					if (child.getName().equalsIgnoreCase(Term.AND)){
						boolean andClash = false;
						TermList andChildren = child.getChildren();
						for (int j=0; j<andChildren.size();j++){
							//tab.log.out ("[BCP] [AND] Checking with " + andChildren.getTerm(j));
							if (ass.contains(andChildren.getTerm(j).negate().simplify())){
								andClash = true;
								break;
							}
						}
						if (andClash){
							//tab.log.out ("[BCP] [AND] Yea, "+child+" clashes! I'm going to remove it");
							li.remove();
						}
					}
				}
				
				// if, at the end of the process, the OR has no children, it means every
				// one of them was clashing... so the whole OR CLASHES! Set it to BOTTOM
				if (t.getChildren().size() == 0) {
					//tab.log.out ("[BCP] Assert has no more children: all of them clash, so the whole assert CLASHES");
					lj.set(a.spawn(Term.BOTTOM));
				}

				// if, instead, the OR has one children, keep it and delete the OR
				if (t.getChildren().size() == 1){
					//tab.log.out ("[BCP] Assert has one child, I'm gonna simplify it");
					if (ass.contains(t.getChild(0))){
						// if the node contains the assert yet just delete it 
						lj.remove();
					}else{
						lj.set(a.spawn(t.getChild(0)));
					}
				}
				//tab.log.out ("[BCP] Old term now is: " + t + "\n");
			}
		}
		Config.core.timers.stopTimer("[TAB][BCP]");
	}

	protected void applyAndRule(Node n){
		// NOTE: this method could make the current node clash!
		Config.core.timers.startTimer("[TAB][AND]");
		if (DEBUG) tab.log.out ("[AND] Applying AND rule on node " + n.getName());

		ListIterator li = n.getAsserts().listIterator();
		while (li.hasNext()){
			// for each assertion
			Assertion ass = (Assertion)li.next();
			Term t = ass.getTerm();
			// if the assertion has the form and(a,b,c)
			if (t.isComplex() && t.getName().equalsIgnoreCase(Term.AND)){
				// remove it
				li.remove();
				TermList children = t.getChildren();
				// and add a, b, c
				for (int i=0; i<children.size(); i++){
					Term child = children.getTerm(i);
					// add assertion only if it's not in the node yet
					// NOTE: the new assertions will have the same 
					// label of their father
					if (!n.contains(child))	li.add(ass.spawn(child));
				}
			}
		}
		Config.core.timers.stopTimer("[TAB][AND]");
	}
}
