/*
 * Created on Mar 22, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jodie.strategies;

import java.util.LinkedList;
import java.util.ListIterator;

import jodie.core.Config;
import jodie.kb.Term;
import jodie.kb.TermList;
import jodie.tableaux.AssertList;
import jodie.tableaux.Assertion;
import jodie.tableaux.Node;
import jodie.tableaux.NodeList;
import jodie.tableaux.Tableaux;

/**
 * @author eynard
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class StrategyALC extends Strategy{

	// inherited protected attributes:
	// Tableaux tab;
	// boolean DEBUG
	
	// inherited protected methods:
	// applyAllRule 
	
	public StrategyALC() {
	}

	public void applyRules(){
		if (tab == null) {
			System.out.println ("Trying to run strategy without a tableaux. Use setTableaux!");
			System.exit(0);
		}
		while (tab.getQueue().size()>0){

			// get the first element from the queue
			// we will pop it away only if there are no more operations to accomplish
			Node n = tab.getQueue().getFirst();

			// first, check the trivial case (node clashes before any rule)
			if (n.clashes()){
				if (DEBUG) tab.log.out ("Found a clash immediately in node " + n.getName());
				tab.setClash();
				return;
			}
			
			// apply rules
			applyUnfoldRule(n);
			if (n.clashes()){
				if (DEBUG) tab.log.out ("Found a clash after unfolding " + n.getName());
				tab.setClash();
				return;
			}
			if (DEBUG) tab.log.out ("Node after UNFOLD is " + n);

			applyAndRule(n);
			if (n.clashes()){
				if (DEBUG) tab.log.out ("Found a clash after AND in node " + n.getName());
				tab.setClash();
				return;
			}
			if (DEBUG) tab.log.out ("Node after AND is " + n);

			applyBCP(n);
			if (DEBUG) tab.log.out ("Node after BCP is " + n);

			applyOrRule(n);
			if (n.clashes()){
				if (DEBUG) tab.log.out ("Found a clash after OR in node " + n.getName());
				tab.setClash();
				return;
			}

			applySomeRule(n);
			
			applyAllRule(n);
			if (trigRules(n)) continue;
			
			tab.getQueue().removeFirst();
		}
	}

	public boolean trigRules(Node n){
		boolean trigger = false;
		AssertList asserts = n.getAsserts();
		
		for (int i=0; i<asserts.size(); i++){
			
			Term t = asserts.getAssert(i).getTerm();
			if (	t.isComplex() && (
					t.getName().equalsIgnoreCase(Term.AND) ||
					t.getName().equalsIgnoreCase(Term.SOME) ||
					t.getName().equalsIgnoreCase(Term.ALL) ||
					t.getName().equalsIgnoreCase(Term.OR) 
					)){
				trigger = true;
				break;
			}
		}
		return trigger;
	}
	
	public void applySomeRule(Node n){
		// NOTE: this method should not create immediate clashing nodes:
		// - first of all, it's always a new node (so no conflicts with ex. ones)
		// - then, the new term is pushed in the queue so if it clashes we'll 
		//   see that later

		Config.core.timers.startTimer("[TAB][SOME]");

		if (DEBUG) tab.log.out ("[SOME] Applying SOME rule on node " + n.getName());

		ListIterator li = n.getAsserts().listIterator();
		while (li.hasNext()){
			// for each assertion
			Assertion ass = (Assertion)li.next();
			Term t = ass.getTerm();
			// if the assertion has the form some(role,concept)
			if (t.isComplex() && t.getName().equalsIgnoreCase(Term.SOME)){

				if (DEBUG) tab.log.out ("[SOME] found " + t);
				// remove it
				li.remove();
				// first term should be a simple role
				Term role = t.getChild(0);
				// second term should be a generic concept
				Term concept = t.getChild(1);

				// NOTE: first check if a node connected through Role...
				NodeList nl = n.getChildren(role.getName()); 

				//       ... and asserting Concept exists yet
				boolean found = false;
				for (int i=0;i<nl.size();i++){
					Node child = (Node)nl.get(i);
					if (child.contains(concept)){
						found = true;
						break;
					}
				}

				if (!found){
					String nodeName = n.getTableaux().getNewAnon();
					if (DEBUG) tab.log.out ("[SOME] creating new node for " + t + "(name: "+ nodeName +")");
					// if not, create a new node with an anonymous name
					Node child = new Node(nodeName,n.getTableaux());
					// make it the child of the current node,
					// through an edge with role's name
					n.addChild(child,role.getName());
					// add the concept to the new node as an assertion
					child.addAssertion(ass.spawn(concept));
					// add the new node to the queue
					n.getTableaux().enqueueNode(child);
				}
			}
		}
		Config.core.timers.stopTimer("[TAB][SOME]");
	}

	public void applyAllRule(Node n){
		// NOTE: this method should not create immediate clashing nodes:
		// - it only add assertions to children nodes, so THEY might clash, not
		//   the current node

		Config.core.timers.startTimer("[TAB][ALL]");

		if (DEBUG) tab.log.out ("[ALL] Applying ALL rule on node " + n.getName());

		ListIterator li = n.getAsserts().listIterator();
		while (li.hasNext()){
			// for each assertion
			Assertion ass = (Assertion)li.next();
			Term t = ass.getTerm();
			// if the assertion has the form all(role,concept)
			if (t.isComplex() && t.getName().equalsIgnoreCase(Term.ALL)){
				// remove it
				li.remove();
				// first term should be a simple role
				Term role = t.getChild(0);
				// second term should be a generic concept
				Term concept = t.getChild(1);
				
				// for all the nodes whose edge has role's name
				LinkedList children = n.getChildren(role.getName());
				for (int i=0;i<children.size();i++){
					// add the concept to the node as an assertion
					Node child = (Node)children.get(i);
					if (!child.contains(concept)){
						// do that only if concept is not asserted yet
						child.addAssertion(ass.spawn(concept));
						// NOTE: since the method has added a new assertion to some 
						//       nodes, we will have to enqueue it for a later check
						n.getTableaux().enqueueNode(child);				
					}
				}
			}
		}
		Config.core.timers.stopTimer("[TAB][ALL]");
	}

	public void applyOrRule(Node n){
		Config.core.timers.startTimer("[TAB][OR]");
		// NOTE: this method could make the current node clash!
		String th = "["+tab.hashCode()+"] ";
		if (DEBUG) tab.log.out ("[OR] "+th+" Applying OR rule on node " + n.getName());

		boolean foundOR = false;
		boolean clashes = true;

		ListIterator li = n.getAsserts().listIterator();
		while (li.hasNext()){
			// for each assertion
			Assertion ass = (Assertion)li.next();
			Term t = ass.getTerm();
			// if the assertion has the form OR(a,b,c)
			if (t.isComplex() && t.getName().equalsIgnoreCase(Term.OR)){
				foundOR = true;
				// remove it
				li.remove();

				if (DEBUG) tab.log.out ("[OR] Working on assertion " + t);

				TermList children = t.getChildren();
				
				// foreach concept...
				for (int i=0; i<children.size(); i++){
					if (DEBUG) tab.log.out ("[OR] " + th + "Trying child number " + i + "(total "+ children.size() +")");
					Term child = children.getTerm(i);

					// from now on, let's try to work on a CLONED tableaux
					n.getTableaux().clearClones();
					Tableaux cloneTab = (Tableaux)n.getTableaux().clone();
					
					// clear the tableaux queue
					cloneTab.getQueue().clear();
					
   					Node cn = n.getTableaux().clonedNode(n); 
					if (cn == null){
						// this should never happen...
						if (DEBUG) tab.log.out ("[OR] ***Haven't found clone of node n!***");
						continue;
					}

					if (DEBUG) tab.log.out ("[OR] Found clone of node n");

					if (DEBUG) tab.log.out ("[OR] Adding assertion " + child + " to node cn");
					// if it's asserted yet...
					if (n.contains(child)) {
						if (DEBUG) tab.log.out ("[OR] Child is asserted yet, I won't add it");
					}else{
						cn.addAssertion(ass.spawn(child));
					}
					
					if (DEBUG) tab.log.out ("[OR] Enqueueing node "+cn.getName()+" to cloneTab");
					cloneTab.enqueueNode(cn);
					
					System.out.println (cn);
					
					if (DEBUG) tab.log.out ("[OR] Applying rulez to cloneTab ("+cloneTab.getQueue()+")");
					cloneTab.applyRules();
					
					//if (!cloneTab.clashes()){
					if (!cloneTab.isClashing()){
						if (DEBUG) tab.log.out ("[OR] Child doesn't clash. Node clashes = " + n.clashes() +
								" clonetab.clashes = " + cloneTab.clashes() + " clonetab.isclashing = "+cloneTab.isClashing());
						if (!n.contains(child)) li.add(ass.spawn(child));
						clashes = false;
						break;
					}
					if (DEBUG) tab.log.out ("[OR] Child clashes!");
				}
				break;
			}
		}
		if (foundOR && clashes){
			if (DEBUG) tab.log.out ("[OR] All children clash, I'll set this node to CLASHES!");
			n.setClash(true);
			//if (DEBUG) tab.log.out ("[OR] Node "+ n.getName() +" clashes = "+ n.clashes());
		}
		Config.core.timers.stopTimer("[TAB][OR]");
	}

}
