/*
 * Created on Mar 22, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jodie.strategies;

import java.util.LinkedList;
import java.util.ListIterator;

import jodie.core.Config;
import jodie.kb.Term;
import jodie.tableaux.Assertion;
import jodie.tableaux.Edge;
import jodie.tableaux.Node;
import jodie.tableaux.NodeList;

/**
 * @author eynard
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class StrategySHI extends StrategySI{

	// inherited protected attributes:
	// Tableaux tab;
	// boolean DEBUG
	
	// inherited protected methods:
	// applyAllRule 
	
	public StrategySHI() {
	}

	/// SHI needs to redefine ALL+ rule and R-neighbour
	
	public void applyAllRule(Node n){
		// NOTE: this method should not create immediate clashing nodes:
		// - it only add assertions to children nodes, so THEY might clash, not
		//   the current node

		Config.core.timers.startTimer("[TAB][ALL]");
		if (DEBUG) tab.log.out ("[ALL] Applying ALL rule on node " + n.getName());

		// if n is indirectly blocked do nothing
		if (n.isIndirectlyBlocked()) return;

		ListIterator li = n.getAsserts().listIterator();
		while (li.hasNext()){
			// for each assertion
			Assertion ass = (Assertion)li.next();
			Term t = ass.getTerm();
			// if the assertion has the form all(role,concept)
			if (t.isComplex() && t.getName().equalsIgnoreCase(Term.ALL)){
				// remove it
				li.remove();
				
				// first term should be a simple role
				Term role = t.getChild(0);
				// second term should be a generic concept
				Term concept = t.getChild(1);
				
				// for all the nodes whose edge has role's name
				LinkedList neighbours = getRNeighbours(n,role.getName());
				for (int i=0;i<neighbours.size();i++){
					// add the concept to the node as an assertion
					Node neighbour = (Node)neighbours.get(i);
					if (!neighbour.contains(concept)){
						// do that only if concept is not asserted yet
						neighbour.addAssertion(ass.spawn(concept));
						// NOTE: since the method has added a new assertion to some 
						//       nodes, we will have to enqueue it for a later check
						n.getTableaux().enqueueNode(neighbour);
					}

					// if the role is transitive
					// NOTE: THIS IS THE CODE FOR THE ALL+ RULE!!!
					if (this.tab.getKB().getTBox().getRole(role.getName()).isTransitive()){
						// add the ALL assertion to the R-Neighbour too
						if (!neighbour.contains(t)){
							// do that only if concept is not asserted yet
							neighbour.addAssertion(ass.spawn(t));
							// NOTE: since the method has added a new assertion to some 
							//       nodes, we will have to enqueue it for a later check
							n.getTableaux().enqueueNode(neighbour);
						}
					}
				}
			}
		}
		Config.core.timers.stopTimer("[TAB][ALL]");
	}

	public NodeList getRNeighbours(Node n, String edgeName){
		// a node y is called an R-Neighbour of x if either
		// y is a successor of x and L(<x,y>)=R or
		// y is a predecessor of x and L(<y,x>)=inv(R)
		NodeList nodes = new NodeList();
		
		// get the name of the inverse role (if it exists)
		String inverseEdgeName = 
			tab.getKB().getTBox().getRole(edgeName).getInverse().getName();

		// get neighbours
		// getEdges returns all the edges starting from n
		for (int i=0; i<n.getEdges().size();i++){
			Edge e = (Edge)n.getEdges().get(i);
			// either you get the edgeName-successor...
			if (e.getName().equalsIgnoreCase(edgeName)){
				nodes.add(e.getBottom());
			}
		}

		Edge e = n.getParentEdge();
		if (e != null && inverseEdgeName != null){
			// ... or you get the inverseEdgeName-predecessor 
			if (e.getName().equalsIgnoreCase(inverseEdgeName)){
				nodes.add(e.getTop());
			}
		}
			
		return nodes;
	}
}
