/*
 * Created on Mar 9, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jodie.tableaux;

import java.util.HashMap;
import java.util.LinkedList;

import jodie.kb.Term;
import jodie.kb.TermList;

/**
 * @author eynard
 */

public class Node implements Cloneable {

	private Tableaux tableaux; // every node knows the tableaux it comes from
	private String name;
	private AssertList asserts = new AssertList();
	private EdgeList edges = new EdgeList();
	private Node parentNode;
	private Edge parentEdge;
	private boolean clashed = false;
	private boolean satisfiable = false;
	
	public Node (String name, Tableaux tableaux){
		this.name = name;
		this.tableaux = tableaux;
	}

	public Object clone(HashMap cloned, Tableaux tab){
		if (cloned.containsKey(this)) {
			//System.out.println("Cloning cached node");
			return cloned.get(this);
		}

		Node n = null;
		try {
			n = (Node)super.clone();

			// assign new tableaux value
			n.tableaux = tab;
			
			// asserts are cloned normally  
			n.asserts = (AssertList)asserts.clone(); // these objs perform deep copy
			
			n.edges = (EdgeList)edges.clone(cloned, tab);
			
			// put the cloned node inside the hash
			cloned.put(this,n);

			//System.out.println ("Node " + n.hashCode() + "-" + n.tableaux.hashCode());
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return n;
	}

	public void cloneFix(Edge parentEdge){

		// fix parentEdge and parentNode
		this.parentEdge = parentEdge;
		if (parentEdge == null){
			this.parentNode = null;
		}else{
			this.parentNode = parentEdge.getTop();
		}
		
		// fix edges
		this.edges.cloneFix(this);
		
	}
	
	public String getName() {
		return name;
	}

	public Tableaux getTableaux(){
		return tableaux;
	}

	public EdgeList getEdges(){
		return edges;
	}
	
	public Edge getParentEdge(){
		return parentEdge;
	}
	
	// getAncestors returns all the ancestors of
	// the current node, inside a NodeList
	public NodeList getAncestors(boolean reflexive){
		NodeList ancestors = new NodeList();
		Node current = this; 
		if (reflexive){ 
			// "ancestors" is reflexive
			ancestors.add(current);
		}
		while (current.parentNode != null){
			ancestors.add(current.parentNode);
			current = current.parentNode;
		}
		return ancestors;
	}
	
	// this getAncestors version is reflexive by default 
	public NodeList getAncestors(){
		return getAncestors(true);
	}
	
	public boolean isSatisfiable(){
		return this.satisfiable;
	}
	
	public void setSatisfiable(){
		this.satisfiable = true;
	}

	public void setSatisfiable(boolean sat){
		this.satisfiable = sat;
	}
	
	public void setName(String name) {
		this.name = name;
	}

	public void setParentEdge(Edge parentEdge) {
		this.parentEdge = parentEdge;
	}

	public void setParentNode(Node parentNode) {
		this.parentNode = parentNode;
	}

	public Node getParentNode() {
		return this.parentNode;
	}

	public void addAssertions (AssertList ass){
		for (int i=0;i<ass.size();i++){
			addAssertion(ass.getAssert(i));
		}
	}

	public void addAssertion (Assertion ass){
		// NOTE: we don't want to work on the object we are passed!
		//       We just want to COPY the assertion to this node.
		// So, let's create a new assertion, with the same term
		Assertion a = new Assertion(ass.getTerm());
		asserts.add(a);
	}

	public void setAssertion (Assertion ass){
		asserts.clear();
		Assertion a = new Assertion(ass.getTerm());
		asserts.add(a);
	}

	public AssertList getAsserts(){
		return asserts;
	}
	
	public void setClash (boolean clash){
		this.clashed = clash;
	}
	
	public Edge addEdge(String edgeName){
		Edge e = new Edge(edgeName);
		e.setTop(this);
		this.edges.add(e);
		return e;
	}
	
	public void addChild (Node node, String edgeName){
		Edge e = addEdge(edgeName);
		node.setParentNode(this);
		node.setParentEdge(e);
		e.setBottom(node);
	}

	public NodeList getChildren(){
		return getChildren(null);
	}

	// NOTE: getchildren works fine as getRNeighbours while roles
	//       don't have inverse: we create getRNeighbours to handle
	//       this case
	public NodeList getChildren(String edgeName){
		NodeList nodes = new NodeList();
		for (int i=0; i<edges.size();i++){
			Edge e = (Edge)edges.get(i);
			if (edgeName==null || e.getName().equalsIgnoreCase(edgeName))
				nodes.add(e.getBottom());
		}
		return nodes;
	}
	
	public NodeList getUSuccessors(){
		NodeList nodes = new NodeList();
		// first, add this node
		nodes.add(this);
		
		// get all the U-children, recursively
		NodeList children = getChildren(Edge.noLabel);
		for (int i=0; i<children.size(); i++){
			// foreach child, add it and its successors
			nodes.addAll(children.getNode(i).getUSuccessors());
		}
		return nodes;
	}

	public boolean contains(Term t){
		return asserts.contains(t);
	}

	public boolean contains(Term t, int idx){
		return asserts.contains(t,idx);
	}
	
	public boolean preTableaux(){
		boolean pt = true;

		// System.out.println ("*** Checking preTab for node "+ this);

		// if the node has no assertions, then it's a PT
		if (this.getAsserts().size() == 0){
			return true;
		}

		// first of all, node has to be clash-free
		if (this.clashes()){
			pt = false;
		}else{
			// the node is clash-free, now check that it 
			// doesn't contain unexpanded conj or disj

			AssertList asserts = this.getAsserts();
			for (int i=0; i<asserts.size(); i++){
				Term child = asserts.getAssert(i).getTerm();

				// first of all, conjunctions
				if (child.isComplex() && child.getName().equalsIgnoreCase(Term.AND)){
					pt = false;
				}
				// then, disjunctions
				else if (child.isComplex() && child.getName().equalsIgnoreCase(Term.OR)){
					boolean expOR = false;
					TermList orChildren = child.getChildren();
					for (int j=0; j<orChildren.size(); j++){
						if (asserts.contains(orChildren.getTerm(j))){
							// if asserts contain at least one of the OR terms,
							// then the OR assertion is considered expanded
							expOR = true;
							break;
						}
					}
					if (!expOR) pt = false; 
				}
				if (!pt) break;
			}
		}
		return pt;
	}
	
	// *** NOTE ***
	// a new version of clashes could return the label of the clashing assert
	public boolean clashes(){
		// trivial case: the node has been marked as clashed yet
		// (this usually happens only in ALC(f) OR rule)
		if (this.clashed) return true;
		
		boolean clash = false;
		
		for (int i=0;i<asserts.size();i++){
			// get a concept
			Term t = asserts.getAssert(i).getTerm();
			// if an assertion is BOTTOM (for instance, a simplified term)
			// then generate a clash
			if (t.equals(Term.BOTTOM)){
				clash = true;
				break;
			}

			// check if there's a clash between asserts, stopping at asserts.size()-1
			if (i<asserts.size()-1 && asserts.contains(t.negate().simplify(),i+1)){
				clash = true;
				break;
			}
		}
		this.clashed = clash; 
		return clash;
	}

	// does a recursive check for clash. It's used to check if any
	// node clashes and determine tableaux clash
	public boolean clashesRec(){
		if (this.clashes()) return true;
		//System.out.println ("Node "+ this.name+ " doesn't clash. Checkin' children");
		boolean clash = false;
		for (int i=0; i<this.getChildren().size();i++){
			if (this.getChildren().getNode(i).clashesRec()){
				//System.out.println ("Node "+ this.getChildren().getNode(i).name+ " clashes. Returning true");
				clash = true;
				break;
			}
		}
		return clash;
	}
	
	// NOTE: valutare l'ipotesi di spostare i metodi sul blocking nella strategia
	// a node x is blocked if, for some ancestor y, y is blocked or L(x)=L(y)
	public boolean isBlocked(){
		// if rootnode return false
		if (this.getParentNode()==null) return false;

		boolean isB = false;
		int i = 0;
		
		// get ancestors list: in this case, ancestors has not
		// to be considered reflexive, otherwise l(x)=l(y) foreach x=y
		NodeList ancestors = this.getAncestors(false);
		
		while (i<ancestors.size() && isB == false){
			// get ancestor
			Node ancestor = ancestors.getNode(i);
			
			if (!this.getAsserts().equals(ancestor.getAsserts())){
				// if L(y) = L(x) then x is blocked
				isB = true;
			}else{
				// otherwise return y.isBlocked()
				isB = ancestor.isBlocked();
			}
			
			i++;
		}
		return isB;
	}
	
	// a blocked node x is indirectly blocked if its predecessor node
	// is blocked, otherwise it's directly blocked
	public boolean isIndirectlyBlocked(){
		// if rootnode return false
		if (this.getParentNode()==null) return false;
		// else...
		return this.getParentNode().isBlocked();
	}
	
	public String toString(){
		String s="";
		s += "Node " + name;
		if (this.clashed) s+= " (*CLASHED*)";
		if (this.isSatisfiable()) s+= "(SAT)";
		s += "\n";
		s += "     ParentNode: " + (parentNode==null?"none":parentNode.getName()) + "\n";
		s += "     ParentEdge: " + (parentEdge==null?"none":parentEdge.getName()) + "\n";
		for (int i=0;i<asserts.size();i++)
			s += "     Assert: " + asserts.getAssert(i) + "\n";
		LinkedList children = getChildren();
		for (int i=0;i<children.size();i++){
			Node child = (Node)children.get(i);
			s += "     Child: " + child.getName() + 
                 " [" + child.parentEdge.getName()  +"]\n";
		}

		for (int i=0;i<children.size();i++){
			Node child = (Node)children.get(i);
			s += child.toString();
		}
		return s;
	}
	
	public String toXML(){
		String s="";
		s += "<node name=\"" + name + "\" pnode=\"" + 
		       (parentNode==null?"none":parentNode.getName()) + "\" pedge=\"" +
			   (parentEdge==null?"none":parentEdge.getName()) + "\" clashes=\"" +
			   (clashed?"Yes":"No") + "\" satisfiable=\"" +
			   (isSatisfiable()?"Yes":"No") +  "\" preTableaux=\"" +
			   (preTableaux()?"Yes":"No") + "\">\n";
		for (int i=0;i<asserts.size();i++)
			s += "<assert>" + asserts.getAssert(i) + "</assert>\n";
		
		LinkedList children = getChildren();
		if (children.size()>0){
			s += "<children>\n";
			for (int i=0;i<children.size();i++){
				Node child = (Node)children.get(i);
				s += child.toXML();
			}
			s += "</children>\n";
		}
		s += "</node>\n";
		return s;
	}
	
	public String toDOT(){
		String s="";
		s += this.name + " [label=\"" + this.name + "\"";
		if (this.clashes()) s += " shape = doublecircle";
		s += "];\n";

		LinkedList children = getChildren();
		for (int i=0;i<children.size();i++){
			Node child = (Node)children.get(i);
			s += child.toDOT();
			s += this.name + "->" + child.name + "[label=\"" + child.parentEdge.getName() + "\"];\n";
		}
		
		return s;
	}
}
