/*
 * Created on Apr 11, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jodie.tableaux;

import java.io.File;
import java.io.FileWriter;
import java.util.HashMap;
import java.util.LinkedList;

import jodie.core.Config;
import jodie.core.Log;
import jodie.kb.KnowledgeBase;
import jodie.strategies.Strategy;

/**
 * @author eynard
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class Tableaux implements Cloneable{
	
	private KnowledgeBase kb;

	private static final String anonPrefix = "anon";
	private int anonNum = 0;

	private Queue queue = new Queue();
	private Node rootNode;
	
	private LinkedList labelstack = new LinkedList();
	private HashMap labels = new HashMap();
	private HashMap cloned = new HashMap(); //keeps track of cloned nodes

	private boolean clash = false;

	public final Log log = new Log(Config.log_tab_name,Config.log_tab_type); 
	
	public Tableaux(KnowledgeBase kb){
		this.kb = kb;
		rootNode = new Node(anonPrefix + String.valueOf(anonNum++),this);
		enqueueNode(rootNode);
	}

	public void setClash(){
		this.clash = true;
	}

	public boolean clashes(){
		return this.clash;
	}

	public boolean isClashing(){
		// this is a more "dynamic" version of clashes:
		// it checks every time if there's a clashing node
		return this.rootNode.clashesRec();
	}
	
	public Queue getQueue(){
		return queue;
	}
	
	public KnowledgeBase getKB(){
		return this.kb;
	}

	public void clearClones(){
		cloned.clear();
	}
	
	public void putLabel(Label l){
		labels.put(l.getHash(),l);
	} 
	
	public Label getLabel(String hash){
		return (Label)labels.get(hash);
	}

	public void pushLabel(Label l){
		labelstack.addFirst(l.getHash());
		putLabel(l);
	}
	
	public Label popLabel(){
		return getLabel((String)labelstack.removeFirst());
	}
	
	public Object clone(){
		// the clone and the labels way are different choices, so we don't clone labels
		Config.core.timers.startTimer("Cloning");
		Tableaux tab = null;
		
		try {
			tab = (Tableaux)super.clone();
			//System.out.println ("I have cloned tableaux " + this.hashCode() + " to " + tab.hashCode());
			// attributes
			// anonPrefix should remain constant, so it's not a problem
			// anonNum SHOULD be cloned ok

			// I can skip the null test for queue and rootNode 'cause they are never null
			tab.rootNode = (Node)this.rootNode.clone(cloned, tab);
			tab.queue = (Queue)this.queue.clone(cloned, tab);
			
			tab.cloneFix();
			
			// empty the cloned cache for the new tableaux
			// (NOTE: the cloned cache will contain the cloned elements 
			//        of the new tableaux, whenever I decide to clone it) 
			tab.cloned = new HashMap();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		Config.core.timers.stopTimer("Cloning");
		return tab;
	}

	public void cloneFix(){
		this.rootNode.cloneFix(null);
	}
	
	public Node clonedNode(Node orig){
		return (Node)cloned.get(orig);
	}
	
	public String getNewAnon(){
		return anonPrefix + String.valueOf(anonNum++);
	}

	public Node getRootNode() {
		return rootNode;
	}

	public void enqueueNode(Node n){
		// if the node is not in the queue yet, enqueue it
		if (!queue.contains(n)) queue.add(n);
	}

	public void applyRules(){
		// get the completion strategy
		Strategy s = Config.core.getKB().getStrategy(this);
		// apply rulez
		s.applyRules();
	}
	
	public String toString(){
		String s = "";
		s += "*** Tableaux ***\n" + queue.toString() + "\n";
		s += "Tableaux " + (this.clashes()?"clashes":"doesn't clash") + "\n";
		s += rootNode.toString();
		
		return s;
	}
	
	public String toXML(){
		return rootNode.toXML();
	}
	
	public String toDOT(){
		return rootNode.toDOT();		
	}
	
	public void toPNG(){
		if (!Config.hasGV) return;
		try {
			FileWriter fw = new FileWriter(Config.dotFileName);
			fw.write("digraph LO{\nsize=\"100,6\";\nnode[shape=circle];\n" + toDOT() + "}");
			fw.close();
			File dir = new File (Config.dotHomeDir);
			Process p = Runtime.getRuntime().exec(Config.dotCommand,null,dir);
			// NOTE: you could add some code to check if dot behaves as we want
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

}
