// The MIT License
//
// Copyright (c) 2003 Ron Alford, Mike Grove, Bijan Parsia, Evren Sirin
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

package org.mindswap.pellet.utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.mindswap.pellet.output.TableData;

/**
 * 
 * @author Evren Sirin
 */
public class Timers  {
	private Map timers = new HashMap();
	
	public Timer mainTimer;
	
	public Timers() {
	    mainTimer = createTimer("main");
	    mainTimer.start();
	}

	public Timer createTimer(String name) {
		Timer t = new Timer(name, mainTimer);
		timers.put(name, t);
		return t;
	}

	public Timer startTimer(String name) {
		Timer t = getTimer(name);
		if(t == null) t = createTimer(name);
		t.start();
		return t;
	}

	public void checkTimer(String name) {
		Timer t = getTimer(name);
		if (t == null)
			throw new UnsupportedOperationException("Timer " + name + " does not exist!");
		
		t.check();
	}
	
	public void resetTimer(String name) {
		Timer t = getTimer(name);
		if (t == null)
		    throw new UnsupportedOperationException("Timer " + name + " does not exist!");
		
		t.start();		
	}
	
	public void setTimeout(String name, long timeout) {
		Timer t = getTimer(name);
		if (t == null)
			t = createTimer(name);
		t.setTimeout(timeout);
	}
		
	public void stopTimer(String name) {
		Timer t = getTimer(name);
		if (t == null)
		    throw new UnsupportedOperationException("Timer " + name + " does not exist!");
		
		t.stop();
	}

	public void resetAll() {
	    Iterator i = timers.values().iterator();
	    while(i.hasNext()) {
	        Timer timer = (Timer) i.next();
	        timer.reset();
	    }
	}
	
	public Timer getTimer(String name) {
		return (Timer) timers.get(name);
	}
	
	public void print() {
		String[] colNames = {"Name", "Count", "Avg", "Total" };
		List list = new ArrayList(timers.values());
		Collections.sort(list, new Comparator() {
			public int compare(Object o1, Object o2) {
				return (int)(((Timer)o2).getTotal() - ((Timer)o1).getTotal());
			}			
		});		
		
		TableData table = new TableData(Arrays.asList(colNames));
		Iterator i = list.iterator();
		while(i.hasNext()) {
			Timer timer = (Timer) i.next();
			List row = new ArrayList();
			row.add(timer.getName());
			row.add(new Long(timer.getCount()));
			row.add(new Double(timer.getAverage()));
			row.add(new Long(timer.getTotal()));
			table.addRow(row);
		}
		
		table.print(System.out);
	}
	
	public String toString() {
		return timers.values().toString();
	}
}
